# -*- coding: utf-8 -*-

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2018, Le Tuan Anh <tuananh.ke@gmail.com>"
__credits__ = []
__license__ = "MIT License"
__description__ = "a Python library for Managing, annotating, and converting natural language corpuses using popular formats (CoNLL, ELAN, Praat, CSV, JSON, SQLite, VTT, Audacity, TTL, TIG, ISF)"
__url__ = "https://github.com/neocl/speach/"
__maintainer__ = "Le Tuan Anh"
__version_major__ = "0.1"  # follow PEP-0440
__version__ = "{}a1".format(__version_major__)
__version_long__ = "{} - Alpha 1".format(__version_major__)
__status__ = "Prototype"
